/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


/*****************************************************************************
*
*  MODULE :             List Management
*  SCOPE :              Implements list management
*  FILE :               list.cc
*  AUTHORS :            E. Abecassis
*  CREATION DATE :      August 1, 1993
*  COPYRIGHTS :        (C) Eric Abecassis, 1994
*  HISTORY :
*
******************************************************************************/

/* IMPORTS *******************************************************************/

#ifdef __GNUG__
#pragma implementation "listgen.hxx"
#endif

#ifndef list_HXX
#include "list.hxx"
#endif

#ifdef TRACE
#undef TRACE
#define TRACE(x) x
#else
#undef TRACE
#define TRACE(x)
#endif

/*=
Module Description
This module implements manipulation of simple link list with iterator.
=*/

const __YOOC_LIST_CTOR YOOC_LIST_CTOR = 0 ;

/*= INTERNAL FUNCTION =======================================================*/
void List::insert(void* el, int pos) 
/*=
[DESCRIPTION :]
  Insert an element after position 'pos'.
[DIAGNOSTIC :]
=*/
{
  if ( (pos >= this->size) && (this->size > 0) ) {
    TRACE(cerr << "==> List : Indice out of boundaries" << endl) ;
    return ;
  }

  ListCell *cell, *prev ;
  cell = new ListCell(el) ;

  if (pos > 0) {
    prev = this->get(pos-1) ;
    cell->next = prev->next ;
    prev->next = cell ;
  } else if (pos == 0) {
    cell->next = this->first ;
    this->first = cell ;
  } else if (pos == -1) {
    cell->next = NULL ;
    if (size == 0) {
      this->first = cell ;
    } else {
      prev = this->get(size-1) ;
      prev->next = cell ;
    }
  }

  this->size++ ;
}


/*= INTERNAL FUNCTION ========================================================*/
void* List::remove(int pos)
/*=
[DESCRIPTION :]
  Remove the element which is in position pos
[DIAGNOSTIC :]
=*/
{
  if (pos >= this->size) {
    TRACE(cerr << "==> List : indice out of boundaries" << endl) ;
    return NULL ;
  }

  ListCell *cell, *prev ;

  if (pos > 0) {
    prev = this->get(pos-1) ;
    cell = prev->next ;
    prev->next = cell->next ;
  } else {
    cell = this->first ;
    this->first = cell->next ;
  }

  void* el = cell->elem ;
  delete cell ;
  this->size-- ;
    
  return el ; 
}


/*= INTERNAL FUNCTION =======================================================*/
ListCell* List::get(int pos) const
/*=
[DESCRIPTION :]
  Return the element which is in position pos
[DIAGNOSTIC :]
=*/
{
  if (pos >= this->size) {
    TRACE(cerr << "==> List : indice out of boundaries" << endl) ;
    return NULL ;    
  }

  int cpos = 0 ;
  for (ListCell* curr = this->first; curr != NULL; curr = curr->next) {
    if (cpos++ == pos) {
      break ;
    } 
  }
  return curr ;
}


/*= INTERNAL FUNCTION =======================================================*/
void List::capture(List& l) 
/*=
[DESCRIPTION :]
  Move every element from the receiver to the parameter.
  After this call, the receiver is an empty list
[DIAGNOSTIC :]
=*/
{
  this->first = l.first ;
  this->size = l.size ;
  l.first = NULL ;
  l.size = 0 ;
}

  
/*= INTERNAL FUNCTION =======================================================*/
IterList::IterList(List& l, int min, int max) : myList(&l) 
/*=
[DESCRIPTION :]
  Build an iterator for list l between position min and max
[DIAGNOSTIC :]
=*/
{
  this->reset(min, max) ;
}


/*= INTERNAL FUNCTION =======================================================*/
void IterList::next() 
/*=
[DESCRIPTION :]
  Move the cursor of the iteration to the next element.
  If there is no more element to scan, the current element is set to null
[DIAGNOSTIC :]
=*/
{
  cpos++ ;
  if (cpos <= end) {
    this->prev = this->curr ;
    this->curr = this->curr->next ;
  } else {
    this->curr = NULL ;
  }
}


/*= INTERNAL FUNCTION =======================================================*/
void* IterList::remove() 
/*=
[DESCRIPTION :]
  Remove the current element of the scan. The cursor is moved to the next
  element. During a scan, 'remove' and 'next' move the cursor of two positions.
  So, if you want to remove elements during scan, its better to used a while
  loop than a for loop.
[DIAGNOSTIC :]
=*/
{
  this->end-- ;
  this->myList->size-- ;
  void* el = this->curr->elem ;

  if (this->prev != NULL) {
    this->prev->next = this->curr->next ;
  } else {
    this->myList->first = this->curr->next ;
  }

  if (this->cpos > this->end) {
    this->curr = NULL ;
  } else {
    this->curr = this->curr->next ;
  }

  return el ;
}


/*= INTERNAL FUNCTION =======================================================*/
void IterList::insert(void* el) 
/*=
[DESCRIPTION :]
  Insert after the current element of the scan. The cursor is moved to the new
  element.
[DIAGNOSTIC :]
=*/
{
  this->cpos++ ;
  this->end++ ;
  this->myList->size++ ;

  if (this->curr != NULL) {
    ListCell* cell = new ListCell(el) ;
    cell->next = this->curr->next ;
    this->curr->next = cell ;
    this->curr = cell ;
  } else {
    this->myList->first = new ListCell(el) ;
    this->curr = this->myList->first ;
  }
}


/*= INTERNAL FUNCTION =======================================================*/
void IterList::reset(int min, int max) 
/*=
[DESCRIPTION :]
  Reset the iteration for a new iteration between min and max
[DIAGNOSTIC :]
=*/
{
  this->start = min ;
  this->cpos = min ;
  if (max == -1) {
    this->end = this->myList->count() - 1 ;
  } else { 
    this->end = max >= this->myList->count() ? this->myList->count()-1 : max ;
  }

  if (this->cpos == 0) {
    this->prev = NULL ;
    this->curr = this->myList->first ;
  } else {
    this->prev = this->myList->get(this->cpos-1) ;
    this->curr = this->prev->next ;
  }
}


